# filestore: Dmedia hashing protocol and file layout
# Copyright (C) 2011 Novacut Inc
#
# This file is part of `filestore`.
#
# `filestore` is free software: you can redistribute it and/or modify it under
# the terms of the GNU Affero General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option) any
# later version.
#
# `filestore` is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU Affero General Public License for more
# details.
#
# You should have received a copy of the GNU Affero General Public License along
# with `filestore`.  If not, see <http://www.gnu.org/licenses/>.
#
# Authors:
#   Jason Gerard DeRose <jderose@novacut.com>

"""
`filestore` - Dmedia hashing protocol and file layout.

This is the heart of Dmedia.  It's been split out of Dmedia to make it easier
to review, and in hopes that other apps might use the hashing protocol and
`FileStore` even if they don't use the higher level Dmedia features.

The best way to experiment is to use a `TempFileStore`.  For example:

>>> from filestore.misc import TempFileStore
>>> fs = TempFileStore()
>>> tmp_fp = fs.allocate_tmp()
>>> tmp_fp.write(b'Hello, Bruce Schneier!')
22
>>> ch = fs.hash_and_move(tmp_fp)
>>> ch.id
'Y685HWMJEE5J39SLBEP4Y3WY7D8Y9JCIQYAFUFQ39MMV84EW'
>>> ch.file_size
22

Learn more about Dmedia here:

    https://launchpad.net/dmedia


Request For Comment
===================

We're seeking feedback from the crypto community on the Dmedia Hashing Protocol
before declaring it final.

In particular, we'd love your thoughts on:

    1. Potential drawbacks to using a hash-list

    2. How we cryptographically tie the root-hash to the file-size

    3. Whether the 240-bit Skein-512 digest size is too small, too large, or
       just right

    4. Whether tying the leaf-hash to the leaf-index (leaf position) is useful
       enough to justify the (small) increase in protocol complexity

Hashing protocol: http://docs.novacut.com/filestore/protocol.html

Or you can build the docs in the source tree if you have python-sphinx
installed:

    cd doc/
    make html
    firefox _build/html/index.html
"""

import os
from os import path
import io
import stat
import fcntl
import json
from threading import Thread
from queue import Queue
from collections import namedtuple
import logging
import time

from dbase32 import db32enc, isdb32, random_id, DB32ALPHABET, RANDOM_B32LEN

from .protocols import TYPE_ERROR
from .protocols import VERSION1 as PROTOCOL

try:
    from _filestore import fallocate, fastread
except ImportError:
    def fallocate(fd, mode, offset, length):
        os.ftruncate(fd, length)

    def fastread(fd, index, size):
        return os.read(fd, size)


__version__ = '14.09.0'

log = logging.getLogger('filestore')

LEAF_SIZE = PROTOCOL.leaf_size
DIGEST_BITS = PROTOCOL.digest_bits
DIGEST_BYTES = PROTOCOL.digest_bytes
DIGEST_B32LEN = PROTOCOL.digest_b32len

# Handy constants for file layout:
DOTNAME = '.dmedia'

# Max durability we allow a local, physical location to count for.  This is
# conservatively at 2 so that a single physical location is never enough (as
# Dmedia will try to maintain at least 3 copies).
MAX_COPIES = 2

# V1 Dbase32 subdir names:
DB32NAMES = tuple(a + b for a in DB32ALPHABET for b in DB32ALPHABET)

# Named tuples:
ContentHash = namedtuple('ContentHash', 'id file_size leaf_hashes')
File = namedtuple('File', 'name size mtime')
Stat = namedtuple('Stat', 'id name size mtime')
Leaf = namedtuple('Leaf', 'index data')
Batch = namedtuple('Batch', 'files size count')
BatchItem = namedtuple('BatchItem', 'file content_hash')
StatVFS = namedtuple('StatVFS', 'size used avail readonly frsize ')

# Special communication constants used by batch_reader()
EndFile = 'EndFile'
EndBatch = 'EndBatch'

IO_READABLE = (io.BufferedReader, io.BufferedRandom)

# For the README.txt in .dmedia directories:
README = """This directory is automatically managed by Dmedia.

Warning: if you make manual changes in this directory, it will quite likely
result in data loss!

This is a control directory for Dmedia.  It contains files and directories
arranged in the special FileStore layout.  To learn more about Dmedia, see:

    https://launchpad.net/dmedia
"""


##############################
# The Dmedia Hashing Protocol:

def hash_leaf(leaf_index, leaf_data):
    """
    Return the leaf-hash of the leaf at *leaf_index* with *leaf_data*.

    For example:

    >>> db32enc(hash_leaf(2, b'XYZ'))
    'YLP5A3SDEEP69SSTFRK7J98AUBS7SW4E75RQN9D4HKRCUXWS'

    :param leaf_index: an ``int`` >= 0
    :param leaf_data: optional ``bytes`` instance with contents of this leaf
    """
    return PROTOCOL.hash_leaf(leaf_index, leaf_data)


def hash_root(file_size, leaf_hashes):
    """
    Return the root-hash for a file *file_size* bytes with *leaf_hashes*.

    For example:

    >>> hash_root(31415, b'NNNNNNNNNNNNNNNNNNNNNNNNNNNNNN')
    'YFXKI7P3BFRMDBBLL4UDVHK3QTHGSXKU4VQ3SPJNKF96TBPV'

    :param file_size: an ``int`` >= 1
    :param leaf_hashes: a ``bytes`` instance that is the concatenated leaf
        hashes produced by `hash_leaf()`
    """
    return db32enc(PROTOCOL.hash_root(file_size, leaf_hashes))


class Hasher:
    """
    A helper to keep track of state as you hash leaf after leaf.
    """

    __slots__ = ('protocol', 'enc', 'file_size', 'leaf_index', 'array', 'closed')

    def __init__(self, protocol=PROTOCOL, enc=db32enc):
        self.protocol = protocol
        self.enc = enc
        self.file_size = 0
        self.leaf_index = 0
        self.array = bytearray()
        self.closed = False

    def hash_leaf(self, leaf):
        if not isinstance(leaf, Leaf):
            raise TypeError(TYPE_ERROR.format('leaf', Leaf, type(leaf), leaf))
        if self.closed:
            raise Exception('Cannot call Hasher.hash_leaf() when Hasher.closed')
        if leaf.index != self.leaf_index:
            raise Exception('Expected leaf.index {}, got {}'.format(
                    self.leaf_index, leaf.index)
            )
        if len(leaf.data) < LEAF_SIZE:
            self.closed = True
        leaf_hash = self.protocol.hash_leaf(leaf.index, leaf.data)
        self.array.extend(leaf_hash)
        self.file_size += len(leaf.data)
        self.leaf_index += 1
        return leaf_hash

    def content_hash(self):
        self.closed = True
        leaf_hashes = bytes(self.array)
        return ContentHash(
            self.enc(self.protocol.hash_root(self.file_size, leaf_hashes)),
            self.file_size,
            leaf_hashes
        )



####################
# Custom exceptions:

class IDError(Exception):
    """
    Raised by `check_id()` when a file ID is not well-formed.

    Note that this says nothing cryptographic about the id, just that either the
    length is incorrect or it contains characters not in B32ALPHABET.
    """
    def __init__(self, _id):
        self.id = _id
        super().__init__('invalid file ID: {!r}'.format(_id))


class LeafHashesError(Exception):
    """
    Raised by `check_leaf_hashes()` when *leaf_hashes* is not well-formed.
    """

    def __init__(self, leaf_hashes):
        self.leaf_hashes = leaf_hashes
        super().__init__('invalid leaf-hashes')


class RootHashError(Exception):
    """
    Raised by `check_root_hash()` when computed ID does not equal expected ID.

    This is exception has ``id``, ``file_size``, and ``leaf_hashes`` attributes,
    and a ``bad_id`` attribute with the root-hash computed from the supplied
    *file_size* and *leaf_hashes*.
    """
    def __init__(self, _id, file_size, leaf_hashes, bad_id):
        self.id = _id
        self.file_size = file_size
        self.leaf_hashes = leaf_hashes
        self.bad_id = bad_id
        super().__init__('{!r}'.format(_id))


class PathError(Exception):
    """
    Raised by `check_path()` when a path is relative or non-normalized.
    """
    def __init__(self, untrusted, abspath):
        self.untrusted = untrusted
        self.abspath = abspath
        super().__init__('{!r} resolves to {!r}'.format(untrusted, abspath))


class PathTraversal(Exception):
    """
    Raised when a path is outside of `FileStore.basedir`.
    """
    def __init__(self, untrusted, abspath, store):
        self.untrusted = untrusted
        self.abspath = abspath
        self.store = store
        super().__init__('{!r} outside of {!r}'.format(abspath, store.basedir))


class NestedParent(Exception):
    """
    Raised when `FileStore.parentdir` is nested in another filestore.
    """
    def __init__(self, parentdir, dotname):
        self.parentdir = parentdir
        self.dotname = dotname
        super().__init__('{!r} contains {!r}'.format(parentdir, dotname))


class CorruptFile(Exception):
    """
    Raise when a file is found to have data corruption.
    """
    def __init__(self, _id, store, **kw):
        self.id = _id
        self.store = store
        self.kw = kw
        super().__init__('{!r} is corrupted in {!r}'.format(_id, store))


class FileNotFound(Exception):
    def __init__(self, _id, store):
        self.id = _id
        self.store = store
        super().__init__('{!r} not in {!r}'.format(_id, store))



#############################
# Input validation functions:

def check_id(_id):
    """
    Verify that *_id* is a valid Dbase32 encoded ID of the correct length.

    A malicious *_id* could cause path traversal or other security gotchas,
    thus this sanity check.  When *_id* is valid, it is returned unchanged:

    >>> check_id('39AY39AY39AY39AY39AY39AY39AY39AY39AY39AY39AY39AY')
    '39AY39AY39AY39AY39AY39AY39AY39AY39AY39AY39AY39AY'

    However, if *_id* is the incorrect length, an `IDError` is raised:

    >>> check_id('39AY39AY39AY39AY39AY39AY')
    Traceback (most recent call last):
      ...
    filestore.IDError: invalid file ID: '39AY39AY39AY39AY39AY39AY'

    Likewise when *_id* is the correct length but contains symbols not included
    in the Dbase32 alphabet:

    >>> check_id('29AZ29AZ29AZ29AZ29AZ29AZ29AZ29AZ29AZ29AZ29AZ29AZ')
    Traceback (most recent call last):
      ...
    filestore.IDError: invalid file ID: '29AZ29AZ29AZ29AZ29AZ29AZ29AZ29AZ29AZ29AZ29AZ29AZ'

    """
    if not isinstance(_id, str):
        raise TypeError(TYPE_ERROR.format('_id', str, type(_id), _id))
    if not (len(_id) == DIGEST_B32LEN and isdb32(_id)):
        raise IDError(_id)
    return _id


def check_leaf_hashes(leaf_hashes):
    """
    Verify that *leaf_hashes* is a ``bytes`` instance of correct length.

    If *leaf_hashes* is valid, it is returned unchanged:

    >>> check_leaf_hashes(b'Y' * 30)
    b'YYYYYYYYYYYYYYYYYYYYYYYYYYYYYY'

    If *leaf_hashes* is invalid, `LeafHashesError` is raised:

    >>> check_leaf_hashes(b'N' * 15)
    Traceback (most recent call last):
      ...
    filestore.LeafHashesError: invalid leaf-hashes

    """
    if not isinstance(leaf_hashes, bytes):
        raise TypeError(TYPE_ERROR.format(
                'leaf_hashes', bytes, type(leaf_hashes), leaf_hashes
            )
        )
    if len(leaf_hashes) == 0 or len(leaf_hashes) % DIGEST_BYTES != 0:
        raise LeafHashesError(leaf_hashes)
    return leaf_hashes


def check_root_hash(_id, file_size, leaf_hashes, unpack=False):
    """
    Verify that claimed *file_size* and *leaf_hashes* are correct for *_id*.

    For example, if the *file_size* was tampered with:

    >>> leaf_hash = hash_leaf(0, b'Hello, Bruce Schneier!')
    >>> _id = hash_root(22, leaf_hash)
    >>> check_root_hash(_id, 21, leaf_hash)  # 21 instead of 22 bytes
    Traceback (most recent call last):
      ...
    filestore.RootHashError: 'Y685HWMJEE5J39SLBEP4Y3WY7D8Y9JCIQYAFUFQ39MMV84EW'

    If the claimed *file_size* and *leaf_hashes* are correct, the 3 values are
    returned in a `ContentHash` named tuple:

    >>> ch = check_root_hash(_id, 22, leaf_hash)
    >>> ch.id
    'Y685HWMJEE5J39SLBEP4Y3WY7D8Y9JCIQYAFUFQ39MMV84EW'
    >>> ch.file_size
    22

    """
    computed = hash_root(file_size, leaf_hashes)
    if _id != computed:
        raise RootHashError(_id, file_size, leaf_hashes, computed)
    if unpack:
        leaf_hashes = tuple(iter_leaf_hashes(leaf_hashes))
    return ContentHash(_id, file_size, leaf_hashes)


def check_path(untrusted):
    """
    Ensure that *untrusted* is a normalized absolute path.

    This is to help protect against path-traversal attacks and to prevent use of
    ambiguous relative paths.

    For example, if *untrusted* is not a normalized absolute path,
    `PathError` is raised:

    >>> check_path('/foo/../root')
    Traceback (most recent call last):
      ...
    filestore.PathError: '/foo/../root' resolves to '/root'

    Otherwise *untrusted* is returned unchanged:

    >>> check_path('/foo/bar')
    '/foo/bar'

    """
    if path.abspath(untrusted) != untrusted:
        raise PathError(untrusted, path.abspath(untrusted))
    return untrusted



#################################################
# Utility functions for working with leaf_hashes:

def enumerate_leaf_hashes(leaf_hashes):
    """
    Enumerate *leaf_hashes*.

    This will yield ``(leaf_index,leaf_hash)`` tuples for each leaf-hash in
    *leaf_hashes*.  For example:

    >>> leaf_hashes = b'AAAAAAAAAAAAAAAAAAAAAAAAAAAAAABBBBBBBBBBBBBBBBBBBBBBBBBBBBBB'
    >>> list(enumerate_leaf_hashes(leaf_hashes))
    [(0, b'AAAAAAAAAAAAAAAAAAAAAAAAAAAAAA'), (1, b'BBBBBBBBBBBBBBBBBBBBBBBBBBBBBB')]

    """
    check_leaf_hashes(leaf_hashes)
    for i in range(len(leaf_hashes) // DIGEST_BYTES):
        yield (i, leaf_hashes[i*DIGEST_BYTES : (i+1)*DIGEST_BYTES])


def iter_leaf_hashes(leaf_hashes):
    """
    Iterate through *leaf_hashes*.

    This will yield a ``leaf_hash`` for each leaf-hash in *leaf_hashes*.
    For example:

    >>> leaf_hashes = b'AAAAAAAAAAAAAAAAAAAAAAAAAAAAAABBBBBBBBBBBBBBBBBBBBBBBBBBBBBB'
    >>> list(iter_leaf_hashes(leaf_hashes))
    [b'AAAAAAAAAAAAAAAAAAAAAAAAAAAAAA', b'BBBBBBBBBBBBBBBBBBBBBBBBBBBBBB']

    """
    check_leaf_hashes(leaf_hashes)
    for i in range(len(leaf_hashes) // DIGEST_BYTES):
        yield leaf_hashes[i*DIGEST_BYTES : (i+1)*DIGEST_BYTES]


def get_leaf_hash(leaf_hashes, i):
    """
    Return a slice containing the i-th leaf_hash from *leaf_hashes*.

    For example:

    >>> lh = b'AAAAAAAAAAAAAAAAAAAAAAAAAAAAAABBBBBBBBBBBBBBBBBBBBBBBBBBBBBB'
    >>> get_leaf_hash(lh, 0)
    b'AAAAAAAAAAAAAAAAAAAAAAAAAAAAAA'
    >>> get_leaf_hash(lh, 1)
    b'BBBBBBBBBBBBBBBBBBBBBBBBBBBBBB'
    >>> get_leaf_hash(lh, 2)
    b''

    """
    start = i * DIGEST_BYTES
    stop = start + DIGEST_BYTES
    return leaf_hashes[start:stop]


def missing_leaves(fp, leaf_hashes):
    """
    Resume a transfer by iterating through leaves that are corrupted or missing.

    This will yield an ``(index,leaf_hash)`` tuple for each leaf that must be
    downloaded.
    """
    fp.seek(0)
    for (i, leaf_hash) in enumerate_leaf_hashes(leaf_hashes):
        data = fp.read(LEAF_SIZE)
        if not data or hash_leaf(i, data) != leaf_hash:
            yield (i, leaf_hash)



###########################################
# High performance IO and batch operations:

def _can_skip(st):
    return st.st_uid == 0 and st.st_gid == 0 and stat.S_IMODE(st.st_mode) == 0o700

def iter_files(basedir):
    """
    Recursively iterate through files in directory *basedir*.

    Note that this does not catch errors like ``OSError``.  We specifically want
    these errors to propagate up!  We don't want a permission error to be
    interpreted as there being no files on the card!

    This function yields a `File` namedtuple for each file found in *basedir*.

    Directory symlinks are not followed, and file symlinks are ignored.
    """
    dirs = []
    for name in sorted(os.listdir(basedir)):
        fullname = path.join(basedir, name)
        st = os.lstat(fullname)
        if stat.S_ISREG(st.st_mode):
            yield File(fullname, st.st_size, st.st_mtime)
        elif stat.S_ISDIR(st.st_mode):
            if name == 'lost+found' and path.ismount(basedir) and _can_skip(st):
                log.warning('Skipping lost+found directory %r', fullname)
                continue
            dirs.append(fullname)
        elif stat.S_ISLNK(st.st_mode):
            log.warning('Ignoring symlink %r', fullname)
    for fullname in dirs:
        for file in iter_files(fullname):
            yield file


def scandir(basedir):
    """
    Recursively scan files in directory *basedir*.

    This function uses `iter_files()` to find files in *basedir*  The result is
    returned as a `Batch` namedtuple.
    """
    check_path(basedir)
    files = tuple(iter_files(basedir))
    return Batch(
        files,
        sum(f.size for f in files),
        len(files),
    )


class SmartQueue(Queue):
    """
    Queue with custom get() that raises exception instances from the queue.
    """

    def get(self, block=True, timeout=None):
        item = super().get(block, timeout)
        if isinstance(item, Exception):
            raise item
        return item


def _start_thread(target, *args):
    thread = Thread(target=target, args=args)
    thread.daemon = True
    thread.start()
    return thread


def reader(src_fp, queue):
    """
    Read leaves from *src_fp* and put them into the *queue*.

    A `Leaf` namedtuple is put into the *queue* for each leaf in *src_fp*.

    After the last leaf has been read, ``None`` is put into the *queue*.

    :param src_fp: A file opened in mode "rb" or "r+b"
    :param queue: A `SmartQueue` instance or similar
    """
    try:
        if not isinstance(src_fp, IO_READABLE):
            raise TypeError(
                TYPE_ERROR.format('src_fp', IO_READABLE, type(src_fp), src_fp)
            )
        src_fp.seek(0)
        fd = src_fp.fileno()
        st = os.fstat(fd)
        if st.st_size > LEAF_SIZE:
            os.posix_fadvise(fd, 0, st.st_size, os.POSIX_FADV_SEQUENTIAL)
        index = 0
        while True:
            data = fastread(fd, index, LEAF_SIZE)
            if not data:
                queue.put(None)
                break
            queue.put(Leaf(index, data))
            index += 1
    except Exception as e:
        log.exception('Error reading %r', src_fp.name)
        queue.put(e)


def reader_iter(src_fp):
    """
    Iterate through leaves in *src_fp*, reading in a separate thread.

    The function yields a `Leaf` namedtuple for each leaf in *src_fp*.

    The reading is done in a separate thread using the `reader()` function.

    This function uses a `SmartQueue` so that if an exception is caught in the
    reader thread, it gets raised in the main thread when `SmartQueue.get()`
    retrieves the exception from the queue.

    :param src_fp: A file opened in mode "rb" or "r+b"
    """
    if not isinstance(src_fp, IO_READABLE):
        raise TypeError(
            TYPE_ERROR.format('src_fp', IO_READABLE, type(src_fp), src_fp)
        )
    q = SmartQueue(4)
    thread = _start_thread(reader, src_fp, q)
    while True:
        leaf = q.get()
        if leaf is None:
            break
        yield leaf
    thread.join()  # Make sure reader() terminates


def batch_reader(batch, queue):
    """
    Read leaves from all files in *batch* and put them into *queue*.

    For each `File` namedtuple in *batch.files*, the `File` namedtuple is put
    into *queue*, marking the start of that file.

    Then a `Leaf` namedtuple is put into *queue* for each leaf in that file.

    After the last leaf is read from the file, the `EndFile` constant is put
    into *queue*, marking the end of that file.

    After the final file is read, the `EndBach` constant is put into *queue*,
    marking the end of the batch.

    :param batch: A `Batch` namedtuple as returned by `scandir()`.
    :param queue: A `SmartQueue` instance or similar
    """
    try:
        for file in batch.files:
            queue.put(file)
            log.info('Opening %r', file)
            src_fp = open(file.name, 'rb')
            fd = src_fp.fileno()
            os.posix_fadvise(fd, 0, file.size, os.POSIX_FADV_SEQUENTIAL)
            index = 0
            while True:
                data = fastread(fd, index, LEAF_SIZE)
                if not data:
                    queue.put(EndFile)
                    break
                queue.put(Leaf(index, data))
                index += 1
        queue.put(EndBatch)
    except Exception as e:
        log.exception('batch.count=%r, batch.size=%r', batch.count, batch.size)
        queue.put(e)


def batch_import_iter(batch, *filestores, callback=None):
    """
    Import the files in *batch* into one or more destination *filestores*.

    This function yields a `BatchItem` namedtuple for each file in *batch*.

    `BatchItem.file` is the `File` namedtuple from *batch.files*.

    For non-empty file, `BatchItem.ch` is the `ContentHash` namedtuple for the
    imported file.

    If a file is empty (zero bytes in size), `BatchItem.ch` will be ``None``.

    Dmedia does not store the zero-byte file, so empty files will not be
    imported.  However, higher-level software will generally want to log each
    empty file (for example, Dmedia accounts for all empty files in the import
    document).

    The reading is done continuously in a separate thread using
    `batch_reader()`.

    This function uses a `SmartQueue` so that if an exception is caught in the
    reader thread, it gets raised in the main thread when `SmartQueue.get()`
    retrieves the exception from the queue.

    :param batch: A `Batch` namedtuple as returned by `scandir()`.
    :param filestores: One or more `FileStore` instances in which to import the
        files in *batch*
    """
    # Validate args:
    if not isinstance(batch, Batch):
        raise TypeError(TYPE_ERROR.format('batch', Batch, type(batch), batch))
    if not filestores:
        raise TypeError(
            'batch_import_iter() takes at least one destination FileStore'
        )
    for (i, fs) in enumerate(filestores):
        if not isinstance(fs, FileStore):
            name = 'filestores[{}]'.format(i)
            raise TypeError(TYPE_ERROR.format(name, FileStore, type(fs), fs))
    if not (callback is None or callable(callback)):
        raise TypeError('callback not callable: {!r}'.format(callback))

    # Okay, time to work:
    if callback is not None:
        count = 0
        size = 0
    q = SmartQueue(8)
    thread = _start_thread(batch_reader, batch, q)
    while True:
        file = q.get()
        if file is EndBatch:
            break
        if file.size == 0:
            if q.get() is not EndFile:
                raise Exception('file.size is 0, but not followed by EndFile')
            if callback is not None:
                count += 1
                callback(count, size)
            yield BatchItem(file, None)
            continue
        temps = tuple(fs.allocate_tmp(file.size) for fs in filestores)
        h = Hasher()
        while True:
            leaf = q.get()
            if leaf is EndFile:
                break
            if callback is not None and leaf.index and leaf.index % 8 == 0:
                callback(count, size + h.file_size)
            h.hash_leaf(leaf)
            for tmp_fp in temps:
                tmp_fp.write(leaf.data)
        ch = h.content_hash()
        if ch.file_size != file.size:
            raise Exception('stat and hasher disagree on file size')
        for (i, fs) in enumerate(filestores):
            tmp_fp = temps[i]
            fs.move_to_canonical(tmp_fp, ch.id)
        if callback is not None:
            count += 1
            size += file.size
            callback(count, size)
        yield BatchItem(file, ch)
    thread.join()  # Make sure batch_reader() terminates


def hash_fp(src_fp, dst_fp=None):
    """
    Compute content hash of open file *src_fp*, optionally writing *dst_fp*.

    This function uses `reader_iter()` to transparently read from *src_fp* in
    another thread.

    The return value is a `ContentHash` named tuple with the ``id``,
    ``file_size``, and ``leaf_hashes``.

    :param src_fp: A file opened in mode "rb" or "r+b"
    :param dst_fp: An optional file opened in mode "wd"
    """
    hasher = Hasher()
    for leaf in reader_iter(src_fp):
        hasher.hash_leaf(leaf)
        if dst_fp:
            dst_fp.write(leaf.data)
    return hasher.content_hash()



###############################################
# Misc functions used by the `FileStore` class:


class Opener:
    __slots__ = ('dir_fd',)

    def __init__(self, dir_fd):
        self.dir_fd = dir_fd

    def __call__(self, name, flags):
        return os.open(name, flags, dir_fd=self.dir_fd)


def _mkdir(name, dir_fd):
    os.mkdir(name, dir_fd=dir_fd)
    os.chmod(name, 0o777, dir_fd=dir_fd)


def _rename(src, dst, dir_fd):
    os.rename(src, dst, src_dir_fd=dir_fd, dst_dir_fd=dir_fd)


def _dumps(obj):
    return json.dumps(obj,
        ensure_ascii=False,
        sort_keys=True,
        separators=(',',': '),
        indent=4,
    )


def _write(name, dir_fd, text):
    fp = open(name, 'x', opener=Opener(dir_fd))
    fp.write(text)
    fp.flush()
    os.fsync(fp.fileno())
    os.chmod(fp.fileno(), 0o444)
    fp.close()


def _create_doc(store_id, copies, **kw):
    if not isinstance(store_id, str):
        raise TypeError(
            TYPE_ERROR.format('store_id', str, type(store_id), store_id)
        )
    if len(store_id) != RANDOM_B32LEN:
        raise ValueError('len(store_id) is {}, need {}: {!r}'.format(
                len(store_id), RANDOM_B32LEN, store_id)
        )
    if not isdb32(store_id):
        raise ValueError('store_id: invalid Dbase32 ID: {!r}'.format(store_id))
    if not isinstance(copies, int):
        raise TypeError(
            TYPE_ERROR.format('copies', int, type(copies), copies)
        )
    if not (0 <= copies <= MAX_COPIES):
        raise ValueError(
            'copies is {}, need 0 <= copies <= {}'.format(copies, MAX_COPIES)
        )
    doc = {
        '_id': store_id,
        'time': time.time(),
        'type': 'dmedia/store',
        'plugin': 'filestore',
        'copies': copies,
    }
    for (key, value) in kw.items():
        doc.setdefault(key, value)
    return doc


def check_doc(doc, expected_id=None):
    if not isinstance(doc, dict):
        raise TypeError(
            TYPE_ERROR.format('doc', dict, type(doc), doc)
        )
    _id = doc['_id']
    if not isinstance(_id, str):
        raise TypeError(
            TYPE_ERROR.format("doc['_id']", str, type(_id), _id)
        )
    if len(_id) != RANDOM_B32LEN:
        raise ValueError("len(doc['_id']) is {}, need {}: {!r}".format(
                len(_id), RANDOM_B32LEN, _id)
        )
    if not isdb32(_id):
        raise ValueError(
            "doc['_id'] is not a valid Dbase32 ID: {!r}".format(_id)
        )
    if not (expected_id is None or expected_id == _id):
        raise ValueError(
            "doc['_id']: expected {!r}; got {!r}".format(expected_id, _id)
        )
    copies = doc['copies']
    if not isinstance(copies, int):
        raise TypeError(
            TYPE_ERROR.format("doc['copies']", int, type(copies), copies)
        )
    if not (0 <= copies <= MAX_COPIES):
        raise ValueError(
            "doc['copies'] is {}, need 0 <= doc['copies'] <= {}".format(
                    copies, MAX_COPIES)
        )
    _type = doc['type']
    if _type != 'dmedia/store':
        raise ValueError(
            "doc['type'] is {!r}, must be 'dmedia/store'".format(_type)
        )
    return doc


def _check_parentdir(parentdir):
    parentdir = check_path(parentdir)
    if DOTNAME in parentdir.split(os.sep):
        raise NestedParent(parentdir, DOTNAME)
    if not path.isdir(parentdir):
        raise ValueError('parentdir not a directory: {!r}'.format(parentdir))
    return parentdir


def _create_filestore(parentdir, store_id, copies, **kw):
    """
    Low level function for creating a Dbase32, V1+ FileStore layout.

    Please use `FileStore.create()` instead of this function

    This is not regarded as part of the public FileStore API and may break
    without warning.
    """
    dir_fd = os.open(parentdir, os.O_DIRECTORY)
    try:
        doc = _create_doc(store_id, copies, **kw)
        tmpname = '-'.join([DOTNAME, store_id])

        # Create all the directories:
        _mkdir(tmpname, dir_fd)
        for name in ('files', 'partial', 'corrupt', 'tmp'):
            _mkdir(path.join(tmpname, name), dir_fd)
        for name in DB32NAMES:
            _mkdir(path.join(tmpname, 'files', name), dir_fd)

        # Write the two files:
        _write(path.join(tmpname, 'README.txt'), dir_fd, README)
        _write(path.join(tmpname, 'filestore.json'), dir_fd, _dumps(doc))

        # Rename to the final directory name, return doc:
        _rename(tmpname, DOTNAME, dir_fd)
        return doc
    finally:
        os.close(dir_fd)


def statvfs(pathname):
    """
    Return usage information for filesystem containing *pathname*.

    This takes the information from ``os.statvfs`` and presents it in a slightly
    higher-level way more fitting for filestore and Dmedia.  A `StatVFS`
    namedtuple is returned with the following attributes:

        * size - the filesystem size in bytes
        * used - space in bytes consumed by existing files
        * avail - space in bytes available to non-privileged users
        * readonly - ``True`` if filesystem is mounted read-only
        * frsize - filesystem block size
    """
    st = os.statvfs(pathname)
    size = st.f_blocks * st.f_frsize
    used = (st.f_blocks - st.f_bfree) * st.f_frsize
    avail = st.f_bavail * st.f_frsize
    readonly = bool(st.f_flag & os.ST_RDONLY)
    return StatVFS(size, used, avail, readonly, st.f_frsize)



########################
# The `FileStore` class:

class FileStore:
    """
    Arranges files in a special layout according to their content-hash.

    Improvements over original `FileStore`:

        1. All directories in layout are created when basedir directory is
           created meaning existence of directories doesn't need to be tested
           for later

        2. Because of (1), `FileStore.path()` and similar no longer require the
           ``create=True`` keyword argument

        3. `FileStore.__init__()` prevents you from creating a `FileStore`
           inside another `FileStore`

        4. Canonical names no longer include the file extension, which
           simplifies the code, eliminates a potential security risk, and allows
           us to provide even stronger data durability guarantees
    """

    dir_fd = None

    @classmethod
    def create(cls, parentdir, store_id=None, copies=1, **kw):
        parentdir = _check_parentdir(parentdir)
        dotdir = path.join(parentdir, DOTNAME)
        if path.exists(dotdir):
            raise ValueError(
                'Cannot create FileStore, dotdir exists: {!r}'.format(dotdir)
            )
        store_id = (random_id() if store_id is None else store_id)
        _create_filestore(parentdir, store_id, copies, **kw)
        return cls(parentdir, store_id)

    def __init__(self, parentdir, expected_id=None):
        """
        Initialize.

        :param parentdir: directory for file store, eg ``'/media/MyDrive'``
        :param excepted_id: if provided, an exception is raised if the ID loaded
            from the ``'filestore.json'`` file doesn't match
        """
        self.parentdir = _check_parentdir(parentdir)
        self.basedir = path.join(self.parentdir, DOTNAME)
        self.dir_fd = os.open(self.basedir, os.O_DIRECTORY |  os.O_NOFOLLOW)
        fcntl.flock(self.dir_fd, fcntl.LOCK_SH | fcntl.LOCK_NB)
        fp = open('filestore.json', 'r', opener=self)
        self.doc = check_doc(json.load(fp), expected_id)
        self.id = self.doc['_id']
        self.copies = self.doc['copies']

        # FIXME: Probably don't need this once we raname relative to dir_fd:
        self.tmp = path.join(self.basedir, 'tmp')

    def __del__(self):
        if self.dir_fd is not None:
            os.close(self.dir_fd)

    def __call__(self, name, flags):
        """
        As a callable, use a `FileStore` as an opener for ``os.open()``.
        """
        return os.open(name, flags, dir_fd=self.dir_fd)

    def __repr__(self):
        return '{}({!r}, {!r})'.format(
            self.__class__.__name__, self.parentdir, self.id
        )

    def __iter__(self):
        """
        Yield a `Stat` namedtuple for each file in this `FileStore`.

        The `Stat` namedtuple has the following attributes:

            * ``id`` - the file's Dmedia ID
            * ``size`` - the file's size in bytes
            * ``mtime`` - the file's modification timestamp

        As the intent of this method is to provide a strong reality check, it's
        rather strict.  It wont yield a `Stat` for any filenames when:

            * the length of the resulting ID isn't DIGEST_B32LEN
            * the resulting ID has characters not in B32ALPHABET
            * the entry is a symlink (even if to a valid file)
            * the file is zero bytes in size
        """
        for prefix in DB32NAMES:
            subdir = path.join(self.basedir, 'files', prefix)
            for name in sorted(os.listdir(subdir)):
                _id = prefix + name
                if len(_id) != DIGEST_B32LEN or not isdb32(_id):
                    continue
                fullname = path.join(subdir, name)            
                st = os.lstat(fullname)
                if stat.S_ISREG(st.st_mode) and st.st_size > 0:
                    yield Stat(_id, fullname, st.st_size, st.st_mtime)

    def _check_dir(self, name):
        assert path.normpath(name) == name
        assert not path.isabs(name)
        st = os.stat(name, dir_fd=self.dir_fd, follow_symlinks=False)
        if not stat.S_ISDIR(st.st_mode):
            raise ValueError(
                '{!r} not a directory in {!r}'.format(name, self.basedir)
            )
        if stat.S_IMODE(st.st_mode) != 0o777:
            raise ValueError(
                'Directory {!r} mode not 0777 in {!r}'.format(name, self.basedir)
            )

    def _check_file(self, name):
        assert path.normpath(name) == name
        assert not path.isabs(name)
        st = os.stat(name, dir_fd=self.dir_fd, follow_symlinks=False)
        if not stat.S_ISREG(st.st_mode):
            raise ValueError(
                '{!r} not a file in {!r}'.format(name, self.basedir)
            )
        if stat.S_IMODE(st.st_mode) != 0o444:
            raise ValueError(
                'File {!r} mode not 0444 in {!r}'.format(name, self.basedir)
            )

    def check_layout(self):
        for name in ('files', 'tmp', 'partial', 'corrupt'):
            self._check_dir(name)
        for name in DB32NAMES:
            self._check_dir(path.join('files', name))
        for name in ('README.txt', 'filestore.json'):
            self._check_file(name)

    def fix_layout_permissions(self):
        # FIXME: This is really a stop-gap to deal with migrating from file
        # layouts created by fairly old versions of FileStore.
        for name in ('files', 'tmp', 'partial', 'corrupt'):
            os.chmod(name, 0o777, dir_fd=self.dir_fd)
        for name in DB32NAMES:
            os.chmod(path.join('files', name), 0o777, dir_fd=self.dir_fd)
        for name in ('README.txt', 'filestore.json'):
            os.chmod(name, 0o444, dir_fd=self.dir_fd)

    def statvfs(self):
        """
        Get disk usage information for the containing file-system.

        For example:

        >>> fs = FileStore('/home/jderose')  #doctest: +SKIP
        >>> fs.statvfs()  #doctest: +SKIP
        StatVFS(size=1936526733312, used=46388051968, avail=1791768784896, readonly=False, frsize=4096)

        This method returns a `StatVFS` namedtuple with the following
        attributes (all sizes are in bytes):

            * size - the filesystem size in bytes
            * used - space in bytes consumed by existing files
            * avail - space in bytes available to non-privileged users
            * readonly - ``True`` if filesystem is mounted read-only
            * frsize - filesystem block size

        Note that ``used`` is the space consumed by all files on the containing
        file-system, not just the files within this `FileStore`.

        Also note that these numbers say nothing about total system storage
        across all mounted file-systems.  Information is returned only for the
        specific file-system upon which this `FileStore` resides.
        """
        return statvfs(self.basedir)

    def catch_traversal(self, untrusted):
        """
        Verify that *untrusted* is inside this `FileStore.basedir` directory.
        """
        abspath = path.abspath(untrusted)
        if abspath.startswith(self.basedir + os.sep):
            return abspath
        raise PathTraversal(untrusted, abspath, self)

    def join(self, *parts):
        """
        Safely join *parts* with basedir directory to prevent path traversal.

        For security reasons, it's very important that you use this method
        rather than ``path.join()`` directly.  This method will prevent path
        traversal attacks, ``path.join()`` will not.

        Also see `FileStore.catch_traversal()`.
        """
        return self.catch_traversal(path.join(self.basedir, *parts))

    def path(self, _id):
        """
        Get the canonical path of a file with *_id*.

        :param _id: base32-encoded content-hash
        """
        _id = check_id(_id)
        return self.join('files', _id[:2], _id[2:])

    def partial_path(self, _id):
        """
        Get the canonical path of a partial file with *_id*.

        "Partial" files are files that are being transfered (either uploaded or
        downloaded).  Because transfers must be resume-able, they are given
        a consistent filename based on their *_id*.

        :param _id: base32-encoded content-hash
        """
        return self.join('partial', check_id(_id))

    def corrupt_path(self, _id):
        """
        Get the canonical path of a corrupt file with *_id*.

        When `FileStore.verify()` finds a file to be corrupted, it moves the
        file into a special holding location based on the file's *_id*.

        The file isn't immediately deleted in case the intact leaves in the file
        are needed to reconstruct a fully valid file.  For example, if you have
        two copies of a file, each which is corrupted, you could still recreate
        the full file as long as the corruption wasn't in the same 8 MiB leaf.

        :param _id: base32-encoded content-hash
        """
        return self.join('corrupt', check_id(_id))

    def random_tmp_path(self):
        """
        Return a random filename within the `FileStore.tmp` directory.

        Note this will not create the file, it will only create a random
        filename. The name will have 120-bits of entropy sourced from
        `os.urandom()` and will be base32-encoded.

        To be clear, a file with the returned filename should *not* exist.

        For safety and security, open the file in mode 'xb' when opening it from
        Python using the `open()` function.

        If this file will be opened by another application, you should try and
        do the same, using the O_CREAT and O_EXCL flags with OPEN(2).
        """
        return path.join(self.tmp, random_id())

    def exists(self, _id):
        """
        Return ``True`` if a file with *_id* exists.

        :param _id: base32-encoded content-hash
        """
        try:
            self.stat(_id)
            return True
        except FileNotFound:
            return False

    def stat(self, _id):
        """
        Return a `Stat` namedtuple for file with *_id*.

        If *_id* is a file in this `FileStore`, a `Stat` namedtuple is retuned
        with the following attributes:

            * ``id`` - the same *_id* you provided
            * ``size`` - the file's size in bytes
            * ``mtime`` - the file's modification timestamp

        If the file is not in this `FileStore`, a `FileNotFound` exception is
        raised.

        As the intent of this method is to provide a strong reality check, it's
        rather strict.  It will also raise `FileNotFound` if:

            * The file exists but is a symlink
            * The file exists but is zero bytes in size
        """
        try:
            fullname = self.path(_id)
            st = os.lstat(fullname)
            if stat.S_ISREG(st.st_mode) and st.st_size > 0:
                return Stat(_id, fullname, st.st_size, st.st_mtime)
        except OSError:
            pass
        raise FileNotFound(_id, self)

    def open(self, _id):
        """
        Open the file with *_id* in ``'rb'`` mode.

        :param _id: base32-encoded content-hash
        """
        return open(self.stat(_id).name, 'rb')

    def verify(self, _id, return_fp=False):
        """
        Verify integrity of file with ID *_id*.

        If the file's root-hash does not equal *_id*, a `FileCorruptFile`
        is raised.

        Otherwise a ContentHash namedtuple is returned with ``id``,
        ``file_size``, and ``leaf_hashes`` attributes.

        Or if called with ``return_fp=True``, the open file is returned after
        calling ``src_fp.seek(0)`` to put read position back at the start of
        the file.

        :param _id: base32-encoded content-hash
        :param return_fp: If ``True``, return open file, if ``False``, return
            a ContentHash namedtuple
        """
        src_fp = self.open(_id)
        c = hash_fp(src_fp)
        if c.id != _id:
            raise self.move_to_corrupt(src_fp, _id, bad_id=c.id)
        if return_fp:
            src_fp.seek(0)
            return src_fp
        return c

    def verify_iter(self, _id, file_size, leaf_hashes):
        """
        Iterate through leaves as they are verified.

        This method will yield ``(leaf_index, leaf_data)`` tuples for each leaf
        in the file identified by *_id*.  This method verifies the ``leaf_data``
        prior to yielding it... so you can trust that the ``leaf_data`` is
        valid by the time you get it.

        The use case that inspired this method was calculating the file md5sum
        prior to uploading it to Amazon S3 using ``boto``.  If the file is
        corrupted, all the md5sum does is make sure the exact same corruption
        arrives intact on S3.

        :param _id: base32-encoded content-hash (an ``str``)
        :param file_size: file size in bytes (an ``int`` > 0)
        :param leaf_hashes: a ``bytes`` instance containing the concatenated
            leaf hashes
        """
        check_root_hash(_id, file_size, leaf_hashes)
        src_fp = self.open(_id)
        st_size = os.fstat(src_fp.fileno()).st_size
        if st_size != file_size:
            raise self.move_to_corrupt(src_fp, _id,
                file_size=file_size,
                bad_file_size=st_size,
            )
        for leaf in reader_iter(src_fp):
            got = hash_leaf(leaf.index, leaf.data)
            expected = get_leaf_hash(leaf_hashes, leaf.index)
            if got != expected:
                raise self.move_to_corrupt(src_fp, _id,
                    leaf_index=leaf.index,
                    leaf_hash=expected,
                    bad_leaf_hash=got,
                )
            yield leaf
        assert get_leaf_hash(leaf_hashes, leaf.index + 1) == b''

    def verify_iter2(self, _id):
        """
        Yield each leaf as it's read, verifying file integrity after final leaf.

        This method will yield a `Leaf` namedtuple for each leaf in the file
        identified by *_id*.  Use this method with care as the integrity of the
        leaves is not known till after the last leaf has been yielded.

        This method is similar to `FileStore.verify_iter()` except only the
        *_id* is needed.
        """
        src_fp = self.open(_id)
        h = Hasher()
        for leaf in reader_iter(src_fp):
            h.hash_leaf(leaf)
            yield leaf
        c = h.content_hash()
        if c.id != _id:
            raise self.move_to_corrupt(src_fp, _id, bad_id=c.id)

    def content_md5(self, _id):
        """
        Compute md5 hash of the file with *_id* for use in Content-MD5 header.
        
        For example:
        
        >>> fs = FileStore('/home/jderose')  #doctest: +SKIP
        >>> fs.content_md5('YDDL5ROVABZP4NBSJPC3HUQDVDAGAP5L26YFXD3UR6N5OLVN')  #doctest: +SKIP
        ('99ca2a74521ad7825768bbfe7fe0dc49', 'mcoqdFIa14JXaLv+f+DcSQ==')

        This method guarantees the correct md5 hash is computed for the file
        with *_id* because it verifies the file as it computes the md5 hash.
        Note that you do not have this guarantee if you simply opened the file
        with `FileStore.open()` and computed the md5 hash that way.

        If you need the md5 hash for use with boto when uploading to S3, use
        this method rather than letting boto compute the md5 hash itself.
        """
        import hashlib
        from base64 import b64encode
        md5 = hashlib.md5()
        for leaf in self.verify_iter2(_id):
            md5.update(leaf.data)
        return (md5.hexdigest(), b64encode(md5.digest()).decode('utf-8'))

    def remove(self, _id):
        """
        Delete file with *_id* from underlying filesystem.

        :param _id: base32-encoded content-hash
        """
        filename = self.path(_id)
        log.info('Removing %r from %r', _id, self)
        os.remove(filename)

    def purge_tmp(self):
        """
        Remove all files in .dmedia/tmp/

        This method is for periodically purging stale temporary files. The
        return value is a list with a `File` namedtuple for each file removed.

        Symlinks are ignored and directories are not traversed, neither of which
        should ever be created by the `FileStore` itself.
        """
        purged = []
        for name in sorted(os.listdir(self.tmp)):
            fullname = path.join(self.tmp, name)
            st = os.lstat(fullname)
            if stat.S_ISREG(st.st_mode):
                purged.append(File(fullname, st.st_size, st.st_mtime))
                log.info('Removing stale temporary file %r', fullname)
                os.remove(fullname)
            else:
                log.info('Unexpected non-regular file %r', fullname)
        return purged

    def allocate_tmp(self, size=None):
        """
        Allocate randomly named temporary file for an import or render.

        When importing a file, supply the file *size* in bytes, and that space
        will be efficiently pre-allocated.

        When rendering, the file size will not be known in advance, so *size*
        should not be supplied.

        :param size: file size in bytes (> 0)
        """
        if not (size is None or isinstance(size, int)):
            raise TypeError(
                TYPE_ERROR.format('size', int, type(size), size)
            )
        if not (size is None or size >= 1):
            raise ValueError(
                'Need size >= 1; got size={}'.format(size)
            )
        filename = self.random_tmp_path()
        tmp_fp = open(filename, 'xb')
        fd = tmp_fp.fileno()
        if size is not None:
            try:
                fallocate(fd, 0, 0, size)
            except OSError:
                os.ftruncate(fd, size)
        os.fchmod(fd, 0o666)
        return tmp_fp

    def allocate_partial(self, size, _id):
        """
        Allocate canonically named partial file for a transfer.

        :param size: file size in bytes (> 0)
        :param _id: base32-encoded content-hash
        """
        if not isinstance(size, int):
            raise TypeError(
                TYPE_ERROR.format('size', int, type(size), size)
            )
        if size < 1:
            raise ValueError(
                'Need size >= 1; got size={}'.format(size)
            )
        filename = self.partial_path(_id)
        try:
            tmp_fp = open(filename, 'rb+')
            fd = tmp_fp.fileno()
            os.ftruncate(fd, size)
        except IOError:
            tmp_fp = open(filename, 'xb')
            fd = tmp_fp.fileno()
            try:
                fallocate(fd, 0, 0, size)
            except OSError:
                os.ftruncate(fd, size)
        os.fchmod(fd, 0o666)
        return tmp_fp

    def move_to_canonical(self, src_fp, _id):
        """
        Move temporary file into its canonical location.

        This method will securely and atomically move a temporary file into its
        canonical location.

        Note, however, that this method does *not* verify the content hash of
        the temporary file!  This is by design as many operations will compute
        the content hash as they write to the temporary file.

        Regardless, the full content hash should have been verified prior to
        calling this method.  To ensure the content is not modified, operations
        must take these steps:

            1. Open *src_fp* and keep it open, thereby retaining a lock on the
               file

            2. Compute the full content hash, which can be done as content is
               written to *src_fp* (open in mode ``'r+b'`` to resume a transfer,
               but hash of previously transfered leaves must still be verified)

            3. With *src_fp* still open, move the temporary file into its
               canonical location using this method.

        As a simple locking mechanism, this method takes an open ``file`` rather
        than a filename, thereby preventing the file from being modified during
        the move.  A ``ValueError`` is raised if *src_fp* is already closed.

        For portability reasons, this method requires that *src_fp* be opened in
        a binary mode: ``'rb'``, ``'wb'``, or ``'rb+'``.  A ``ValueError`` is
        raised if opened in any other mode.

        For security reasons, this method will only move a temporary file
        located within the ``FileStore.basedir`` directory or a subdirectory
        thereof.  If an attempt is made to move a file from outside the store,
        `PathTraversal` is raised.  See `FileStore.catch_traversal()`.

        Just prior to moving the file, a call to ``os.fchmod()`` is made to set
        read-only permissions (0444).  After the move, *src_fp* is closed.

        The return value is the absolute path of the canonical file.

        :param src_fp: a file opened in ``'wb'``, ``'rb'``, or ``'rb+'`` mode
        :param _id: base32-encoded Dmedia content-hash
        """
        if not isinstance(src_fp, io.BufferedIOBase):
            raise TypeError(TYPE_ERROR.format(
                    'src_fp', io.BufferedIOBase, type(src_fp), src_fp
                )
            )
        if src_fp.closed:
            raise ValueError(
                'src_fp is closed, must be open: {!r}'.format(src_fp)
            )
        src = self.catch_traversal(src_fp.name)
        dst = self.path(_id)
        # FIXME: Still need to decide the *new* way of handling duplicates in a
        # FileStore... although the *new* way might be the current way below
        src_fp.flush()
        os.fchmod(src_fp.fileno(), stat.S_IRUSR | stat.S_IRGRP | stat.S_IROTH)
        os.rename(src, dst)
        src_fp.close()
        return dst

    def move_to_corrupt(self, src_fp, _id, **kw):
        """
        Move a file from its canonical location to its corrupt location.

        When a file is found to be corrupt (meaning the computed content hash
        doesn't match the expected content hash), it is moved from its canonical
        location to a special corrupt location.

        For data safety and forensics, corrupt files are not immediately
        deleted.  As files can be verified with 8 MiB granularity, it is
        possible to recover the complete file from two separate corrupt copies
        as long as the corruption is in different 8 MiB leaves.

        This method returns an appropriate `CorruptFile` instance to be
        raised by the calling code.
        
        :param src_fp: a file opened in ``'rb'`` mode
        :param _id: base32-encoded Dmedia content-hash
        """
        if not isinstance(src_fp, io.BufferedReader):
            raise TypeError(TYPE_ERROR.format(
                    'src_fp', io.BufferedReader, type(src_fp), src_fp
                )
            )
        if src_fp.closed:
            raise ValueError(
                'src_fp is closed, must be open: {!r}'.format(src_fp)
            )
        src = self.catch_traversal(src_fp.name)
        dst = self.corrupt_path(_id)
        log.info('Moving corrupt file %r to %r', src, dst)
        os.rename(src, dst)
        src_fp.close()
        return CorruptFile(_id, self, **kw)

    def verify_and_move(self, tmp_fp, _id):
        # New in FileStore 14.05: FileStore.verify_and_move() is now API
        # forgiving in the same way FileStore.hash_and_move() is in that you can
        # pass a *tmp_fp* that is a io.BufferedWriter:
        if isinstance(tmp_fp, io.BufferedWriter):
            tmp_fp.close()
            tmp_fp = open(tmp_fp.name, 'rb')
        if not isinstance(tmp_fp, IO_READABLE):
            raise TypeError(
                TYPE_ERROR.format('tmp_fp', IO_READABLE, type(tmp_fp), tmp_fp)
            )
        if tmp_fp.name != self.partial_path(_id):
            raise ValueError(
                'bad partial_path() for {!r}: {!r}'.format(_id, tmp_fp.name)
            )
        ch = hash_fp(tmp_fp)
        if ch.id != _id:
            raise ValueError(
                'expected {!r}, computed {!r}'.format(_id, ch.id)
            )
        self.move_to_canonical(tmp_fp, _id)
        return ch

    def hash_and_move(self, tmp_fp):
        if isinstance(tmp_fp, io.BufferedWriter):
            tmp_fp.close()
            tmp_fp = open(tmp_fp.name, 'rb')
        if not isinstance(tmp_fp, IO_READABLE):
            raise TypeError(
                TYPE_ERROR.format('tmp_fp', IO_READABLE, type(tmp_fp), tmp_fp)
            )
        assert tmp_fp.name.startswith(path.join(self.basedir, 'tmp') + os.sep)
        ch = hash_fp(tmp_fp)
        self.move_to_canonical(tmp_fp, ch.id)
        return ch

    def import_file(self, src_fp):
        """
        Atomically copy open file *src_fp* into this filestore.

        The method will compute the content-hash of *src_fp* as it copies it to
        a temporary file within this store.  Once the copying is complete, the
        temporary file will be moved to its canonical location using
        `FileStore.move_to_canonical()`.

        This method returns an ``(id, file_size, leaf_hashes)`` namedtuple.

        Note that *src_fp* must have been opened in ``'rb'`` mode.

        :param src_fp: a ``file`` instance created with ``open()``
        """
        size = os.fstat(src_fp.fileno()).st_size
        tmp_fp = self.allocate_tmp(size)
        log.info('Importing file %r into %r', src_fp.name, self)
        c = hash_fp(src_fp, tmp_fp)
        assert c.file_size == size
        self.move_to_canonical(tmp_fp, c.id)
        return c

    def copy(self, _id, *filestores):
        """
        Copy file with *_id* from this FileStore into one or more *filestores*.
        """
        if not filestores:
            raise TypeError(
                '{}.copy() takes at least one destination filestore'.format(
                    self.__class__.__name__
                )
            )
        for (i, fs) in enumerate(filestores):
            if not isinstance(fs, FileStore):
                name = 'filestores[{}]'.format(i)
                raise TypeError(
                    TYPE_ERROR.format(name, FileStore, type(fs), fs)
                )
            if fs.parentdir == self.parentdir:
                name = 'filestores[{}]'.format(i)
                raise ValueError(
                    'wont copy into {}, same parentdir as {!r}'.format(name, self)
                )
        src_fp = self.open(_id)
        size = os.fstat(src_fp.fileno()).st_size
        temps = [fs.allocate_tmp(size) for fs in filestores]
        h = Hasher()
        for leaf in reader_iter(src_fp):
            h.hash_leaf(leaf)
            for tmp_fp in temps:
                tmp_fp.write(leaf.data)
        ch = h.content_hash()
        if ch.id != _id: 
            for tmp_fp in temps:
                tmp_fp.close()
                os.remove(tmp_fp.name)
            raise self.move_to_corrupt(src_fp, _id, bad_id=ch.id)
        assert ch.file_size == size
        for (i, fs) in enumerate(filestores):
            tmp_fp = temps[i]
            fs.move_to_canonical(tmp_fp, _id)
        return ch

    def copy_all(self, *filestores):
        """
        Copy all files in this FileStore into one or more other *filestores*.
        """
        for st in self:
            try:
                yield self.copy(st.id, *filestores)
            except CorruptFile as e:
                yield e

