# filestore: Dmedia hashing protocol and file layout
# Copyright (C) 2013 Novacut Inc
#
# This file is part of `filestore`.
#
# `filestore` is free software: you can redistribute it and/or modify it under
# the terms of the GNU Affero General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option) any
# later version.
#
# `filestore` is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU Affero General Public License for more
# details.
#
# You should have received a copy of the GNU Affero General Public License along
# with `filestore`.  If not, see <http://www.gnu.org/licenses/>.
#
# Authors:
#   Jason Gerard DeRose <jderose@novacut.com>

"""
Migrate a V0 layout to V1, rehash to map V0 to V1 IDs.
"""

import os
from os import path
from copy import deepcopy
import json
import time
import logging

import dbase32
from dbase32 import rfc3548

from . import check_doc, reader_iter, Hasher, DB32NAMES, README
from . import _check_parentdir, _create_doc, _dumps
from .misc import load_data, write_test_vector_files
from .protocols import VERSION0


log = logging.getLogger()

# V0 Base32 subdir names:
B32ALPHABET = '234567ABCDEFGHIJKLMNOPQRSTUVWXYZ'
B32NAMES = tuple(a + b for a in B32ALPHABET for b in B32ALPHABET)

# The presense of any of these means a V0, Base32 file layout:
NAMES_DIFF = tuple(sorted(set(B32NAMES) - set(DB32NAMES)))


def is_v0_files(files):
    for name in NAMES_DIFF:
        if path.isdir(path.join(files, name)):
            return True
    return False


def b32_to_db32(_id):
    """
    Re-encode an ID from Base32 to Dbase32.

    >>> b32_to_db32('5Q2VGOCRKWGDCSJOHDMXYFCE')
    'WJTO9H5KDP965LCHA6FQR857'

    """
    return dbase32.db32enc(rfc3548.b32dec(_id))


def v0_doc_to_v1(old):
    assert isinstance(old, dict)
    new = deepcopy(old)
    new['_id'] = b32_to_db32(old['_id'])
    new['type'] = 'dmedia/store'
    new['plugin'] = 'filestore'
    return check_doc(new)


def open_file(fullname):
    try:
        return open(fullname, 'r')
    except FileNotFoundError:
        pass


def load_doc(fullname):
    try:
        return json.load(open(fullname, 'r'))
    except (FileNotFoundError, ValueError):
        pass


class Migration:
    def __init__(self, parentdir):
        self.parentdir = _check_parentdir(parentdir)
        self.basedir = path.join(self.parentdir, '.dmedia')
        self.files = self.join('files')
        self.files0 = self.join('files0')
        self.store = self.join('store.json')
        self.store0 = self.join('store0.json')
        self.filestore = self.join('filestore.json')

    def __iter__(self):
        if not self.needs_migration():
            return
        for prefix in B32NAMES:
            subdir = path.join(self.files0, prefix)
            for name in sorted(os.listdir(subdir)):
                src = path.join(subdir, name)     
                v0_id = prefix + name       
                assert path.isfile(src) or path.islink(src)
                assert rfc3548.isb32(v0_id) and len(v0_id) == 48

                if path.islink(src):
                    log.info('Reading symlink %r', src)
                    yield (v0_id, os.readlink(src), None)
                else:
                    src_fp = open(src, 'rb')
                    h0 = Hasher(protocol=VERSION0, enc=rfc3548.b32enc)
                    h1 = Hasher()
                    for leaf in reader_iter(src_fp):
                        h0.hash_leaf(leaf)
                        h1.hash_leaf(leaf)
                    ch0 = h0.content_hash()
                    ch1 = h1.content_hash()
                    assert dbase32.isdb32(ch1.id)
                    if ch0.id != v0_id:
                        yield (v0_id, None, None)
                    else:
                        dst = self.v1_path(ch1.id)
                        log.info('Moving %r to %r', src, dst)
                        os.rename(src, dst)
                        os.symlink(ch1.id, src)
                        yield (v0_id, ch1.id, ch1)

    def needs_migration(self):
        if not path.isdir(self.basedir):
            return False
        self.write_v1_doc_if_needed()
        if self.move_files_if_needed():
            return True
        return path.isdir(self.files0)

    def build_v0_simulation(self):
        os.mkdir(self.basedir)
        os.chmod(self.basedir, 0o777)
        for name in ('partial', 'corrupt', 'tmp'):
            self.mkdir(name)
        self.create_v0_files()
        self.write('README.txt', README)
        doc = {
            '_id': rfc3548.random_id(),
            'copies': 1, 
            'plugin': 'filestore', 
            'time': time.time(), 
            'type': 'dmedia/store'
        }
        self.write('store.json', _dumps(doc))
        V0 = load_data('V0')
        write_test_vector_files(self.basedir, VERSION0)
        for (key, _id) in V0['root_hashes'].items():
            src = self.join(key)
            dst = self.path(_id)
            os.rename(src, dst)
            os.chmod(dst, 0o444)
        return doc

    def join(self, *parts):
        if not parts:
            raise ValueError('Cannot have empty parts')
        return path.join(self.basedir, *parts)

    def mkdir(self, *parts):
        fullname = self.join(*parts)
        os.mkdir(fullname)
        os.chmod(fullname, 0o777)
        return fullname

    def write(self, name, text):
        fullname = self.join(name)
        fp = open(fullname, 'x')
        fp.write(text)
        fp.flush()
        os.fsync(fp.fileno())
        os.chmod(fp.fileno(), 0o444)
        fp.close()
        return fullname

    def load_v0_doc(self):
        # Hack: as we switched from 'store.json' to 'filestore.json' after the
        # initial migration functionality landed in 13.05, the old V0 doc might
        # be in the 'store0.json' file:
        fp = open_file(self.store0) or open_file(self.store)
        if fp is None:
            return
        try:
            return json.load(fp)
        except ValueError:
            pass        

    def get_v1_doc(self):
        # If store0.json is presest, it should always be a V0, Base32 doc:
        old = load_doc(self.store0)
        if old is not None:
            return v0_doc_to_v1(old)

        # If only store.json is presest, there are a few scenarios:
        old = load_doc(self.store)
        if old is not None:
            old_id = old['_id']
            # This scenario is always clear-cut:
            if not dbase32.isdb32(old_id):
                return v0_doc_to_v1(old)
            # Also clear-cut, but 'migration' flag only in 13.05 daily builds:
            if old.get('migrated') is True or not rfc3548.isb32(old_id):
                old['type'] = 'dmedia/store'
                old['plugin'] = 'filestore'
                old.pop('migrated', None)
                return check_doc(old)
            # High chance of being correct, but possible for it to be wrong:
            return v0_doc_to_v1(old)

        # Otherwise create a new V1 doc, without any migration:
        return _create_doc(dbase32.random_id(), 1)

    def write_v1_doc_if_needed(self):
        if path.isfile(self.filestore):
            return False
        doc = self.get_v1_doc()
        self.write('filestore.json', _dumps(doc))
        assert path.isfile(self.filestore)
        return True

    def path(self, _id):
        assert isinstance(_id, str)
        assert rfc3548.isb32(_id) or dbase32.isdb32(_id)
        assert len(_id) == 48
        return self.join('files', _id[:2], _id[2:])

    def v0_path(self, v0_id):
        assert isinstance(v0_id, str)
        assert rfc3548.isb32(v0_id)
        assert len(v0_id) == 48
        return self.join('files0', v0_id[:2], v0_id[2:])

    def v1_path(self, v1_id):
        assert isinstance(v1_id, str)
        assert dbase32.isdb32(v1_id)
        assert len(v1_id) == 48
        return self.join('files', v1_id[:2], v1_id[2:])

    def create_v0_files(self):
        """
        For testing or reversing the migration, create the V0 layout:
        """
        self.mkdir('files')
        for name in B32NAMES:
            self.mkdir('files', name)

    def create_v1_files(self):
        """
        After moving a V0 files to files0, we need to create a correct V1 files.
        """
        self.mkdir('files')
        for name in DB32NAMES:
            self.mkdir('files', name)

    def move_files_if_needed(self):
        if is_v0_files(self.files):
            self.move_v0_files_to_files0()
            return True
        if not path.isdir(self.files):
            self.create_v1_files()
        return False  

    def move_v0_files_to_files0(self):
        assert is_v0_files(self.files)
        assert not path.exists(self.files0)
        log.warning('Moving V0 files/ to files0/ in %r', self.basedir)
        os.rename(self.files, self.files0)
        self.create_v1_files()
