#!/usr/bin/env python3

# filestore: dmedia hashing protocol and file layout
# Copyright (C) 2011-2014 Novacut Inc
#
# This file is part of `filestore`.
#
# `filestore` is free software: you can redistribute it and/or modify it under
# the terms of the GNU Affero General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option) any
# later version.
#
# `filestore` is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU Affero General Public License for more
# details.
#
# You should have received a copy of the GNU Affero General Public License along
# with `filestore`.  If not, see <http://www.gnu.org/licenses/>.
#
# Authors:
#   Jason Gerard DeRose <jderose@novacut.com>

"""
Install `filestore`.
"""

import sys
if sys.version_info < (3, 4):
    sys.exit('ERROR: FileStore requires Python 3.4 or newer')

import os
from os import path
import subprocess
from distutils.core import setup, Extension
from distutils.cmd import Command

import filestore
from filestore.tests.run import run_tests


TREE = path.dirname(path.abspath(__file__))


def run_under_same_interpreter(opname, script, args):
    print('\n** running: {}...'.format(script), file=sys.stderr)
    if not os.access(script, os.R_OK | os.X_OK):
        print('ERROR: cannot read and execute: {!r}'.format(script),
            file=sys.stderr
        )
        print('Consider running `setup.py test --skip-{}`'.format(opname),
            file=sys.stderr
        )
        sys.exit(3)
    cmd = [sys.executable, script] + args
    print('check_call:', cmd, file=sys.stderr)
    subprocess.check_call(cmd)
    print('** PASSED: {}\n'.format(script), file=sys.stderr)


def run_pyflakes3():
    script = '/usr/bin/pyflakes3'
    names = [
        'filestore',
        'setup.py',
        'dmediasum',
    ]
    args = [path.join(TREE, name) for name in names]
    run_under_same_interpreter('flakes', script, args)


def run_sphinx_doctest():
    script = '/usr/share/sphinx/scripts/python3/sphinx-build'
    doc = path.join(TREE, 'doc')
    doctest = path.join(TREE, 'doc', '_build', 'doctest')
    args = ['-EW', '-b', 'doctest', doc, doctest]
    run_under_same_interpreter('sphinx', script, args)


class Test(Command):
    description = 'run unit tests and doctests'

    user_options = [
        ('skip-flakes', None, 'do not run pyflakes static checks'),
        ('skip-sphinx', None, 'do not run Sphinx doctests'),
    ]

    def initialize_options(self):
        self.skip_flakes = 0
        self.skip_sphinx = 0

    def finalize_options(self):
        pass

    def run(self):
        if not self.skip_flakes:
            run_pyflakes3()
        if not run_tests():
            sys.exit(2)
        if not self.skip_sphinx:
            run_sphinx_doctest()


_filestore = Extension('_filestore', sources=['_filestore.c'])


setup(
    name='filestore',
    description='dmedia hashing protocol and file layout',
    url='https://launchpad.net/filestore',
    version=filestore.__version__,
    author='Jason Gerard DeRose',
    author_email='jderose@novacut.com',
    license='AGPLv3+',
    packages=[
        'filestore',
        'filestore.tests'
    ],
    package_data={'filestore': ['data/*.json']},
    scripts=['dmediasum'],
    cmdclass={'test': Test},
    ext_modules=[_filestore],
)
